
"""IMAP Sieve client.

Based on RFC 3028


Author: Ulrich Eck <ueck@net-labs.de> April 2001


Public class:		Sieve
"""

__version__ = "0.01"

import binascii, re, socket, string, time, random, sys


#	Globals

CRLF = '\r\n'
SIEVE_PORT = 2000

#	Commands

Commands = {
	# name		  valid states
	'AUTHENTICATE':	('NONAUTH',),
	'LOGOUT':	('NONAUTH', 'AUTH', 'LOGOUT'),
	'GETSCRIPT':	('AUTH', ),
	'PUTSCRIPT':	('AUTH', ),
	'SETACTIVE':	('AUTH', ),
	'DELETESCRIPT':	('AUTH', ),
	'LISTSCRIPTS':	('AUTH', ),
	}

#  regex

ScriptList = re.compile(r'\s*"(?P<Script>.*)"\s*(?P<Active>.*)')
SizeTag = re.compile(r'\{(?P<size>\d*)\}')
Welcome_response = re.compile(r'\s*"(?P<type>[A-Z-]+)"\s*"(?P<data>.*)"')

class SIEVE:

	"""Sieve client class.

	Instantiate with: Sieve([host[, port]])

		host - host's name (default: localhost);
		port - port number (default: standard Sieve port).

	"""

	class error(Exception): pass	# Logical errors - debug required
	class abort(error): pass	# Service errors - close and retry
	class readonly(abort): pass	# Mailbox status changed to READ-ONLY


	def __init__(self, host = '', port = SIEVE_PORT):
		self.host = host
		self.port = port
		self.state = 'NONAUTH'
		self.literal = None
		
		self.capabilities = []
		self.loginmechs = []
		self.implementation = ''

		# Open socket to server.

		self.open(host, port)


		typ,data = self._get_response()
		
		for line in data:
			m = Welcome_response.match(line)
			if m:
				if m.group('type') == "IMPLEMENTATION":
					self.implementation = m.group('data')
				if m.group('type') == "SASL":
					self.loginmechs = string.split(m.group('data'))
				if m.group('type') == "SIEVE":
					self.capabilities = string.split(m.group('data'))


		if typ == 'OK':
			return
		else:
			raise self.error(self.welcome)


	def __getattr__(self, attr):
		#	Allow UPPERCASE variants of SIEVE command methods.
		if Commands.has_key(attr):
			return eval("self.%s" % string.lower(attr))
		raise AttributeError("Unknown SIEVE command: '%s'" % attr)



	#	Public methods


	def open(self, host, port):
		"""Setup 'self.sock' and 'self.file'."""
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.sock.connect((self.host, self.port))
		self.file = self.sock.makefile('r')


	def socket(self):
		"""Return socket instance used to connect to SIEVE server.

		socket = <instance>.socket()
		"""
		return self.sock





	def authenticate(self, mechanism, authobject):
		"""Authenticate command - requires response processing."""
		
		mech = string.upper(mechanism)

		self.literal = binascii.b2a_base64(authobject)[:-1]
		
		val = '"%s" {%d+}' % (mech,len(self.literal))
		option = [self.literal,]
		
		typ, dat = self._command('AUTHENTICATE', val, option)
		if typ != 'OK':
			raise self.error(dat[-1])
		self.state = 'AUTH'
		return typ, dat

	def login(self,user,password):
		""" login to the Sieve server. """
		
		authstr = "%s\0%s\0%s" % (user,user,password)
		
		typ,dat = self.authenticate('PLAIN',authstr)
		
		


	def logout(self):
		"""Shutdown connection to server.

		(typ, [data]) = <instance>.logout()

		Returns server 'BYE' response.
		"""
		self.state = 'LOGOUT'
		try: typ, dat = self._command('LOGOUT')
		except: typ, dat = 'NO', ['%s: %s' % sys.exc_info()[:2]]
		self.file.close()
		self.sock.close()
		return typ,dat






	def listscripts(self):
		"""List Scripts.

		(typ, [data]) = <instance>.listscripts()

		'data' is list of LIST responses.
		"""
		name = 'LISTSCRIPTS'
		typ,dat = self._command(name)
		ret = []
		for l in dat:
			if l[:2] == 'OK' or l[:2] == 'NO':
				break
			m = ScriptList.match(l)
			if m:
				if m.group('Active'):
					ret.append((m.group('Script'),1))
				else:
					ret.append((m.group('Script'),0))
		
		return typ,ret

	def getscript(self,script):
		"""List Scripts.

		(typ, [data]) = <instance>.getscript(scriptname)

		'data' is list of LIST responses.
		"""
		name = 'GETSCRIPT'
		typ,dat = self._command(name,'"%s"' % script)
		if len(dat) > 0:
			m = SizeTag.match(dat[0])
			if m: size = int(m.group('size'))
			else: size = 0
		
		ret = string.join(dat[1:],'\n')
		
		return typ,ret

	def putscript(self,script,data):
		"""List Scripts.

		(typ, [data]) = <instance>.putscript(scriptname,data)

		'data' is list of LIST responses.
		"""
		name = 'PUTSCRIPT'
		value = '"%s" {%d+}' % (script,len(data))
		option = [data,]
		typ,dat = self._command(name,value,option)
		return typ,dat


	def deletescript(self,script):
		"""List Scripts.

		(typ, [data]) = <instance>.deletescript(scriptname)

		'data' is list of LIST responses.
		"""
		name = 'DELETESCRIPT'
		typ,dat = self._command(name,'"%s"' % script)
		return typ,dat

	def setactive(self,script):
		"""List Scripts.

		(typ, [data]) = <instance>.setactive(scriptname)

		'data' is list of LIST responses.
		"""
		name = 'SETACTIVE'
		typ,dat = self._command(name,'"%s"' % script)
		return typ,dat
	

	#	Private methods



	def _get_line(self):

		line = self.file.readline()
 		if not line:
			raise self.abort('socket error: EOF')

		return line[:-1]








	def _command(self, name, value = '', option = []):

		if self.state not in Commands[name]:
			self.literal = None
			raise self.error(
			'command %s illegal in state %s' % (name, self.state))

		if value:
			data = "%s %s" % (name,value)
		else:
			data = name

		try:
			self.sock.send('%s%s' % (data, CRLF))
			print "sent: %s" % data
			for o in option:
				self.sock.send('%s%s' % (o, CRLF))
				print "sent: %s" % o
		except socket.error, val:
			raise self.abort('socket error: %s' % val)

		return self._get_response()
		


	def _get_response(self):
		
		ret = []
		last = ''

		while 1:
			last = self._get_line()
			if string.upper(last[:2]) in ['OK','NO']:
				typ = string.upper(last[:2])
				break 
			ret.append(last)

		if not typ == 'OK':
			return 'NO','%s, %s' % (last,str(ret))
			
		return typ,ret
